/*
 * LCD_Config v2.6
 * ---------------
 *
 * last change: 17.03.2011
 *
 * Allows configuration of the MultiWii software via a serial connection.
 * Requires a LCD display and at least four buttons (left, right, up, down)
 * or an analog joystick.
 *
 * Options:
 *   Logging functionality (requires a SD card), output to binary or LogView OpenFormat.
 *   Status LEDs (three LEDs, two are sufficient without logging)
 *   Setting of LCD backlight intensity
 *   Use of ShiftRegLCD lib instead of LiquidChrystal lib
 *   Hardware buttons instead of buttons connected via cascading resistors
 *   Analog joystick instead of buttons
 *   Graphic output (simple character graphic for angles and servo signals)
 *   MinMax mode (store und display minimum/maximum values of certain fields)
 *   [APC220] (disables a APC220 wireless module during playback of logfile - rather specific option) 
 *   [Test] (for standalone testing without a MWC board connected)
 *
 * IMPORTANT: undefining the LOGGING function does not work correctly. When commenting
 * out the line "#include <SD.h>", the image size decreases by about 4k. Recommended
 * in case you are targeting a ATMega168.
 *
 * ----------
 *
 * Graphic Mode is toggled either by pressing the SELECT key or by pressing the RIGHT key longer
 * than 0.5 sec. Min/Max Mode is toggled by pressing the RIGHT key shorter than 0.5 sec.
 *
 * ----------
 *
 * For infos on LogView fileformat, see www.logview.info.
 * To import logfiles into LogView, you need to copy the attached .ini and .jpg File
 * to the directory <username>/<application data>/LogView/OpenFormat, select the
 * device LCD_CONFIG, and click on "import device file".
 * Data is formatted to three LogView channels: the first channel contains sensor data,
 * the second servo and motor PWM signals, the third the receiver channels. 
 *
 * ----------
 *
 * Authors: yacco, (calli)
 *
 * changes to 1.25: 
 *   display values of gyro and acc changed to signed
 *   show minimum/maximum values of sensor data and servo signals
 * changes to 1.28:
 *   wait state after startup, no serial communication during initialization of the MWC
 *   removed min/max values
 *   added LED status signals
 * changes to 1.30:
 *   bugfix in playback routine, added timeout in playback
 * changes to 1.31:
 *   minor bugfixes, added setting of backlight
 * changes to 1.32:
 *   removed KEY_SELECT column in next_state, backlight and waitForInit optional
 * changes to 1.33:
 *   check trailing 'A' char in serial communication to detect frame errors, 
 *   optimization of format routines (merged two routines, invalid values are displayed as ###)
 *   bugfix in logging (logMode=-1 was stored in flash, so logging was permanently disabled)
 *   KEY_LEFT cancels reading when no serial connection is present (important to reset wrong settings!) 
 * changes to 1.5:
 *   reimplemented Min/Max-Mode (found it useful again), 
 *   fixed compiling problems for certain combinations of options 
 * changes to 1.6:
 *   adaptation to MWC version 1.preter6 (new config "Flying Wing", new data format)
 *   changed status display (state 0 - Config)
 *   bug 'logMode=-1' finally solved
 *   Version has been set to 2.0 to disambiguate between MWC version and LCD_Conf version.
 * changes to 2.0 beta: 
 *   swapped motors 4 and 5 in 'Esc Out' for versions 1.5 and 1.6prebis
 *   changed behaviour of UP and DOWN for consistency (UP decrements, DOWN increments values)
 *   added version V1.6 
 *   output logfiles in LogView format [OPTIONAL] (file numbers starting from 500, extension ".lvw")
 *   added JOYSTICK option to replace input buttons by an analog joystick [experimental]
 *   display 'invalid' values in graphics
 *   heavy optimizations (!) due to space restrictions
 *   removed WAIT_FOR_INIT option, changed start state to 31 instead 
 * changes to 2.0:
 *   bugfix in deserialize() with v1.6
 * changes to 2.01:
 *   Joystick handling was improved (calibration, use of pullups, can be used in parallel to buttons)
 *   Added 'Play LogFile' state
 *   Easter egg function to display internal configuration :)
 *   BUGFIX: RC Rate multiplied by factor 2 for MWC Versions >v1.5, minimum of 20% for RC Rate (0% makes no sense)
 *   Added random offsets in TEST mode
 * changes to 2.1:
 *   Cancel in Logfile Playback (press LEFT_KEY)
 *   Settings of "Baud Rate", serialRead() has been adapted to work with different COM speeds
 *   Greater flexibility in changing the Settings
 *   (Log) Delay Time was moved to Settings state and is used as the time interval of polling the MWC board
 * changes to 2.2:
 *   Changed internal representation of state machine and strings
 *   Support for v1.7pre (I-/P-Level instead of AutoLevelStrength, trigger calibration, 8 instead of 5 receiver
 *   channels, configuration of activation modes for Aux1/Aux2, ...)
 *   Software Version is now stored in Logfile
 *   APC220 Option (disables APC220 module via pin D2 during playback of logfiles, experimental)
 * changes to 2.3:
 *   Swapped Angle and Baro/Comp/Cycle states
 *   Added 'Config' parameter to LogFiles, show Config and Version in selectLogFile
 *   Use of blinking Cursor instead of Arrows in Enter mode
 *   LogView mode: corrected order of receiver channels
 * changes to 2.4:
 *   "Roll/Pitch" state - simultaneous setting of Roll and Pitch PID values
 * changes to 2.5:
 *   Changed representation of configIds
 *   'pos' parameter in lcd output instead of (col, row), saves ~250 bytes of flash
 *   Support for v1.7prebis (replaced v1.7pre)
 *   CHECK_INPUT option to visualize the joystick input (hidden state, press RIGHT in state 0 for >1 sec)
 *   BATTERY_LEVEL option (experimental), visualisation on LCD and via Status LEDS, 4 warn levels can be defined
 *
 * overall footprint (on an Arduino Uno, using the LiquidCrystal lib and cascading buttons): 
 *  ~30574 bytes with all options enabled (far beyond the maximum image size uploadable via bootloader on a AtMega328)
 *  ~13164 bytes with all options disabled (and //#include <SD.h>)
 */

// ----- begin of configuration parameters -----

// use ShiftRegLCD lib instead of LiquidCrystal lib
// [footprint: decreases ~200 bytes]
//#define SRLCD     

// ----------

// use Hardwarekeys connected to Ports
// [footprint: ~120 bytes]
//#define HWKEYS

// Input pins of the hardbuttons: (used when HWKEYS is defined)
#define HW_RIGHT    10
#define HW_UP        9
#define HW_DOWN     11
#define HW_LEFT      8
#define HW_SELECT    6

// Analog Input pin for cascading buttons
#define KEY_PIN      0

// ----------

// Use an analog Joystick (e.g., Sparkfun Thumb Joystick) instead of Keys (experimental!)
// Joystick and Buttons can be used simultaneously. The center position of the Joystick
// is calibrated during startup, so leave the stick in central position when switching on.
// [footprint: ~240 bytes]
#define JOYSTICK

#define NO_JS_SELECT_KEY
#define JS_SELECT_PIN A0
#define JS_HOR_PIN    A1
#define JS_VERT_PIN   A2

// Configuration of the Axes
#define JS_HOR_SIGN  +1 
#define JS_VERT_SIGN -1
#define JS_DEADBAND  48

// inverts the switch level of the SELECT key (default: high)
//#define JS_INV_SELECT_KEY

// ----------

// includes a hidden state to check the joystick input. 
// activated by pressing the RIGHT key in state 0 for more than a second. 
#define CHECK_INPUT

// ----------

// includes code to read and visualize the state of the battery via the OSD functionality
// in MWC Versions >= v1.7prebis
#define BATTERY_LEVEL

#define VBATLEVEL0 111 // 11,1V - green
#define VBATLEVEL1 107 // 10,7V - green/yellow
#define VBATLEVEL2 103 // 10,3V - yellow
#define VBATLEVEL3 99  // 9.9V  - yellow/ red, below: red

// ----------

// include logging to SD/SDHC card (formatted to FAT16 or FAT32) via SPI
// [footprint: ~12 kbytes]
#define LOGGING
#define CHIP_SELECT 10

// enables output in LogView format (~1 kbyte)
#define LOGVIEW

// Logging Modes resp. LogFile Formats:
#define OFFLINE    -2
#define LOG_ERROR  -1
#define NO_LOGGING  0
#define LF_BINARY   1
#define LF_LOGVIEW  2

// ----------

// include routines to show some output as simple character graphic (for servo signals and angles) 
// [footprint: ~2.3 kbytes]
#define GRAPHIC

// ----------

// track minimum and maximum values of certain values
// [footprint: ~650 bytes]
#define MINMAX

#ifdef MINMAX
  #define MM_ACT [0]
  #define MM_DIM [3]
  #define MM_VAL [minMaxMode]
#else
  #define MM_ACT
  #define MM_DIM
  #define MM_VAL
#endif

// ----------

// include routines to show the state of the application with three leds 
// (red: error, yellow: busy/waiting for input, green: logging)
// [footprint: ~220 bytes]
#define STATUS_LEDS

// Caution: Default Pins of LEDs and Joystick input overlap
#define RED_LED      A5
#define YELLOW_LED   A4
#define GREEN_LED    A3

// ----------

// if defined, allows dimming of LCD backlight intensity (Pin 10)
// [footprint: ~50 bytes (?)]
#define BACKLIGHT

#define BACKLIGHT_PIN 10

// -----

// disables a connected APC220 module during playback of logfiles. It is assumed that the ENABLE pin of
// APC220 is connected to pin D2 of the Arduino board. I soldered a 20k resistor between the two pins, 
// seems to work.
// [footprint: 40 bytes]
//#define APC220

#define APC220_ENABLE_PIN 2

// -----

// simulates a serial connection, SELECT key in state "ESC Out" changes the configuration
//#define TEST

// ----------------

// Number of hardbuttons:
#define NUM_KEYS     5

// Identifiers for the five hardbuttons:
#define KEY_RIGHT    0
#define KEY_UP       1
#define KEY_DOWN     2
#define KEY_LEFT     3
#define KEY_SELECT   4
#define STATUS       5 // Last column in the transition table is reserved for status flags

// RC Channel Aliases:
#define ROLL         0
#define PITCH        1
#define YAW          2
#define THROTTLE     3
#define AUX1         4
#define AUX2         5
#define CAM_ROLL     6
#define CAM_PIT      7

// ----- end of configuration parameters -----

// States:
#ifdef LOGGING
  #define NUM_STATES  36
#else
  #define NUM_STATES  35
#endif

// Number of Configurations: (Unknown, Tri, Quad+, QuadX, Bi, Gimbal, Y6, Hex, Flying Wing, Y4, HexX)
#define NUM_CONFIGS  11

// the GUI_SERIAL_SPEED is used for logfile playback only
#define GUI_SERIAL_COM_SPEED 115200

// ----------------

// for persistence of SW Version, Logging Mode, delayTime and Profiles
#include <EEPROM.h>

// defines the PROGMEM macro to move constant variables from SRAM to flash memory
#include <avr/pgmspace.h> 

// Workaround for "insertion of forward declarations in excluded blocks" problem
// Don't move this line around.
void setup();     

#ifndef SRLCD
  #include <LiquidCrystal.h>
  LiquidCrystal lcd(8, 9, 4, 5, 6, 7); // rs, enable, d4, d5, d6, d7
  #define INIT_LCD lcd.begin(16, 2); 
#else
  #include <ShiftRegLCD.h>
  ShiftRegLCD lcd(3, 2, 4, 2);        // (Datapin, Clockpin, Enablepin or TWO_WIRE [, Lines [, Font]]])
  #define INIT_LCD                    // no begin in ShiftRegLCD...
#endif

#ifdef LOGGING
  #include <SD.h>                     // ***** comment this line when LOGGING is not required - saves 4k of flash *****
#endif

// ----------------

static uint8_t backlight;      // backlight intensity (0..255)

#ifdef BACKLIGHT
  #define BL_MSG "Backlight"
  #define NUM_SETTINGS 4
  #define INIT_BACKLIGHT   backlight = 255 - readEEPROM(5); setBacklight();
  void setBacklight()     {analogWrite(BACKLIGHT_PIN, backlight);}
#else
  #define BL_MSG
  #define NUM_SETTINGS 3
  #define INIT_BACKLIGHT
  #define setBacklight()
#endif

// ----------------

PROGMEM prog_char metaData[] = ":" 
#ifdef TEST
  "T"
#endif
#ifdef SRLCD
  "S"
#endif
#ifdef HWKEYS
  "H"
#endif
#ifdef JOYSTICK
  "J"
#endif
#ifdef GRAPHIC
  "G"
#endif
#ifdef MINMAX
  "M"
#endif
#ifdef LOGGING
  "L"
  #ifdef LOGVIEW
  "v"
  #else
  "-"
  #endif
#endif
#ifdef BACKLIGHT
  "B"
#endif
#ifdef STATUS_LEDS
  "$"
#endif
#ifdef BATTERY_LEVEL
  "V"
#endif
  ":"
;

// ----------------

PROGMEM prog_char intro       [] = " LCD Config 2.6";
PROGMEM prog_char emptyLine   [] = "\020";
PROGMEM prog_char reading     [] = "Reading...\006";
PROGMEM prog_char readMsg     [] = " Read\004Bytes\002";
PROGMEM prog_char writeMsg    [] = " Wrote\004Bytes ";
PROGMEM prog_char cancelRead  [] = "Reading canceled";
PROGMEM prog_char frameError  [] = "Inv. FrameLength";
PROGMEM prog_char timeoutError[] = " Timeout Error\002";
PROGMEM prog_char calibrateMsg[] = "[\002Calibrate?\002]";
PROGMEM prog_char yesNoMsg    [] = " \177: no, \176: yes\002";
PROGMEM prog_char delaySuff   [] = "\0100 ms";

PROGMEM prog_char profStates[][ 6] = { " Load", " Save", "Clear", "Reset", }; 
PROGMEM prog_char auxStates [][ 6] = { "Level", "Baro ", "Comp ", "CStab", "CTrig", }; 
PROGMEM prog_char modeMsg   [][ 9] = { "AcroMode", "Lev.Mode", " Offline" };
PROGMEM prog_char setMsgs   [][12] = { "MWC Version", "Delay Time", "Baud Rate", BL_MSG };
PROGMEM prog_char haltMsg     [17] = " Program halted ";

#ifdef BATTERY_LEVEL
  PROGMEM prog_char batLevelMsg [] = "Battery:\006V";
#endif

#ifdef LOGGING
  PROGMEM prog_char logMsg      [] = "Logging...\006";
  PROGMEM prog_char noCard      [] = "No SD Card found";
  PROGMEM prog_char fileError   [] = "Open File failed";
  PROGMEM prog_char deleteFailed[] = "Delete failed\003";
  PROGMEM prog_char haltReqMsg  [] = " Halt Program?\002";
  PROGMEM prog_char logEnbMsg   [] = "Logging enabled ";
  PROGMEM prog_char stopLogMsg  [] = " Stop Logging?\002";
  PROGMEM prog_char noFileMsg   [] = "No Logfile found";
  PROGMEM prog_char fileMsg     [] = "File: ";
//  PROGMEM prog_char logViewType [] = "LV\004";

  PROGMEM prog_char logStates[][15] = { "Start Logging", "Play Logfile", "Delete Logfile", };

  #ifdef LOGVIEW
    PROGMEM prog_char formatMsg[] = "LogView Format?";
  #endif
#endif

/* ----- */

#ifdef MINMAX
  PROGMEM prog_char minMaxDsc[][4] = { "Act", "Min", "Max" };
#endif

/* ----- */

#ifdef TEST
  #define INIT_TEST lcdPrint(12 + 16, "Test"); randomSeed(4711); config = 1;
#else
  #define INIT_TEST
#endif

/* ------ */

PROGMEM prog_char credits[] = { 40, 99, 41, 50, 48, 49, 49, 32, 98, 121, 32, 121, 97, 99, 99, 111, 0 };
int8_t  baudRates        [] = { 1, 2, 4, 8, 12, 16, 32, 48, 96, };
uint8_t defaultProfile   [] = { 'C', 40, 30, 13, 40, 30, 13, 80, 0, 0, 45, 65, 100, 0, 0, 0, };
uint8_t auxMasks         [] = { 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, }; 

// ----------------

#define V1_5       0
#define V1_6_prb   1
#define V1_6       2
#define V1_7_prb   3

PROGMEM prog_char versions[][5] = {
  "1.5 ", 
  "1.6b",   // v1.6prebis
  "1.6 ",
  "1.7b",   // v1.7prebis
};

// number of bytes per frame send from the MultiWii board, dependant on version
uint8_t frameLength[] = { 
  65,  // 1.5
  70,  // 1.6prebis
  74,  // 1.6 (1.6preter)
  83,  // 1.7prebis
};

// configuration ids
PROGMEM prog_char configIds[] = { 
  "Unknown\0" 
  "Tri\0" 
  "Quad+\0" 
  "QuadX\0" 
  "Bi\0"
  "Gimbal\0" 
  "Y6\0" 
  "Hex\0" 
  "Fl.Wing\0"
  "Y4\0"
  "HexX"
};

uint8_t configOffsets [] = { 0, 8, 12, 18, 24, 27, 34, 37, 41, 49, 52, };
char    shortConfigIds[] = "--TrQ+QxBiGmY6H|FwY4H-";

// ----------------

PROGMEM prog_char msg0 [] = "1\002Config\007";
PROGMEM prog_char msg1 [] = "2\002P-I-D";
PROGMEM prog_char msg2 [] = "3\002Rates";
PROGMEM prog_char msg3 [] = "4\002Sensors";
PROGMEM prog_char msg4 [] = "5\002Receiver In";
PROGMEM prog_char msg5 [] = "6\002ESC/Servo Out";
    
PROGMEM prog_char msg38[] = "Roll+Pitch";
PROGMEM prog_char msg6 [] = "Roll";
PROGMEM prog_char msg7 [] = "Pitch";
PROGMEM prog_char msg8 [] = "Yaw";

PROGMEM prog_char msg9 [] = "D/R\003Expo\003ALS";
PROGMEM prog_char msg10[] = "R/P\003Yaw\003Thro";
    
PROGMEM prog_char msg11[] = "Gyr: Rol Pit Yaw";
PROGMEM prog_char msg12[] = "Acc: Rol Pit Z";
PROGMEM prog_char msg13[] = "Alt\002Head\002Cycle";
    
PROGMEM prog_char msg14[] = "Roll\002Pitch\002Yaw";
PROGMEM prog_char msg15[] = "Thro\002Aux1\002Aux2";

PROGMEM prog_char msg16[] = "Lf:\006Ri:";
PROGMEM prog_char msg17[] = "Ru:\006Re:";
    
PROGMEM prog_char msg18[] = "Lf:\006Ri:";
PROGMEM prog_char msg19[] = "Fr:\006Re:";

PROGMEM prog_char msg20[] = "FL:\006FR:";
PROGMEM prog_char msg21[] = "RL:\006RR:";

PROGMEM prog_char msg22[] = "M1:\006M2:";
PROGMEM prog_char msg23[] = "S1:\006S2:";

PROGMEM prog_char msg24[] = "Roll.:";
PROGMEM prog_char msg25[] = "Pitch:";

PROGMEM prog_char msg26[] = "Up";
PROGMEM prog_char msg27[] = "Dn";

PROGMEM prog_char msg28[] = "Fr";
PROGMEM prog_char msg29[] = "Re";
    
PROGMEM prog_char msg30[] = "\002Rho\006Phi";
    
PROGMEM prog_char msg31[] = "7\002Profiles";
PROGMEM prog_char msg32[] = "8\002Settings";

PROGMEM prog_char msg33[] = "S1:\006S2:";
PROGMEM prog_char msg34[] = "S3:\006S4:";

PROGMEM prog_char msg35[] = "Cam.: Roll Pitch";
PROGMEM prog_char msg36[] = "Level";

PROGMEM prog_char msg37[] = "Aux: 1-LMH 2-LMH";

PROGMEM prog_char msg39[] = "Lf";
PROGMEM prog_char msg40[] = "Ri";
PROGMEM prog_char msg41[] = "Up:\006Dn:";

#ifdef LOGGING
PROGMEM prog_char lgMsg[] = "9\002Logging";
#endif

/* ----- */

#define LOG_STATE   35
#define CONF_OFFSET 23

#ifdef LOGGING
  #define UP0 LOG_STATE
  #define DN0 LOG_STATE
  #define LOG_MSG lgMsg,
#else
  #define UP0 8
  #define DN0 0
  #define LOG_MSG
#endif

//State messages
PROGMEM prog_char* msgs[] = {
 msg0,  msg1,  msg2,  msg3,  msg4,  msg5,  msg6,  msg7,  msg8,  msg9, 
 msg10, msg11, msg12, msg13, msg14, msg15, msg16, msg17, msg18, msg19, 
 msg20, msg21, msg22, msg23, msg24, msg25, msg26, msg27, msg28, msg29, 
 msg30, msg31, msg32, msg33, msg34, msg35, msg36, msg37, msg38, msg39, 
 msg40, msg41, LOG_MSG
};

// state transition table, values above 80 are pseudo-states
PROGMEM prog_char nextState[NUM_STATES][NUM_KEYS + 1] = {
  // Right Up  Down Left  Status      Status Flag: -2=Refresh Data
  {  100, UP0,   2,   1,   0,  -2, }, //  0 - Config
  {  100, UP0,   2, 102,   0,  -2, }, //  1 - Acc, Bar, Mag

  {    9,   0,   3,   0,   1,  -1, }, //  2 - PID
  {   14,   2,   4,   0,   2,  -1, }, //  3 - Rates
  {   16,   3,   5,   0,   3,  -1, }, //  4 - Sensors
  {   21,   4,   6,   0,   4,  -1, }, //  5 - Rec In
  {   99,   5,   7,   0,   5,  -1, }, //  6 - ESc Out
  {   88,   6,   8,   0,  31,  -1, }, //  7 - Profile
  {   87,   7, DN0,   0,  32,  -1, }, //  8 - Settings

  {   80, 101,  10,   2,  38,  -1, }, //  9 - PID Roll/Pitch
  {   81,   9,  11,   2,   6,  -1, }, // 10 - PID Roll
  {   82,  10,  12,   2,   7,  -1, }, // 11 - PID Pitch
  {   83,  11, 101,   2,   8,  -1, }, // 12 - PID Yaw
  {   84,  12,   9,   2,  36,  -1, }, // 13 - pLevel, iLevel.

  {   85,  15,  15,   3,   9,  -1, }, // 14 - DR, Expo, AutoLevel Strength
  {   86,  14,  14,   3,  10,  -1, }, // 15 - R/P Rate, Yaw Rate, Throttle PID
 
  {   89,  19,  17,   4,  11,  -2, }, // 16 - Gyro
  {   89,  16,  18,   4,  12,  -2, }, // 17 - Acc
  {   89,  17,  19,   4,  13,  -2, }, // 18 - Angles
  {   89,  18,  16,   4,  30,  -2, }, // 19 - Altitde, Heading, Cycle time
  {   90,  91,  91,   0,  37,  -1, }, // 20 - Activation Modes

  {   89,  23,  22,   5,  14,  -2, }, // 21 - R, P, Y
  {   89,  21,  23,   5,  15,  -2, }, // 22 - Thr, Aux1, Aux2
  {   89,  22,  21,   5,  35,  -2, }, // 23 - Cam Rol, Cam Pit

  {   89,  -1,  -1,   6,  16,  17, }, // 24 - Tri
  {   89,  -1,  -1,   6,  18,  19, }, // 25 - Quad+
  {   89,  -1,  -1,   6,  20,  21, }, // 26 - QuadX
  {   89,  -1,  -1,   6,  22,  23, }, // 27 - Bi
  {   89,  -1,  -1,   6,  24,  25, }, // 28 - Gimbal
  {   89,  -1,  -1,   6,  26,  27, }, // 29 - Y6
  {   89,  -1,  -1,   6,  28,  29, }, // 30 - Hex
  {   89,  -1,  -1,   6,  33,  34, }, // 31 - Flying Wing
  {   89,  -1,  -1,   6,  20,  41, }, // 32 - Y4
  {   89,  -1,  -1,   6,  39,  40, }, // 33 - Hex6X

  {  100, UP0,   2,  20,   0,  -3, }, // 34 - Battery Level

#ifdef LOGGING
  {   95,   8,   0,   0,  42,  -1, }, // 35 - Logging
#endif
  // Right Up  Down Left  Status      Status Flag: -2=Refresh Data
};

// charset used in profile names
static char charsInName[] = " .01234567890_ABCDEFGHIJKLMNOPQRSTUVWXYZ";
static char profileMsg [] = " #00: 01234567  ";
static char statusMsg  [] = "Nck- Baro- Comp-";
static char lineBuffer [17];

// ----------------

#ifdef LOGGING
  static File     logFile               ;
  static char     fileName[] = "000.log";     // LogFiles consists of a three digit number (000..999) with extension ".log"
  static uint16_t lineNo                ;     // current record number in logFile
#endif

static boolean   initialized           ;
static int8_t    state                 ;      // current ui state
static int8_t    profile               ;      // current profile index
static uint8_t   vers                  ;      // MultiWii Software Version 1.5
static uint8_t   baudRate              ;      // baud rate of serial connection to MWC board
static uint8_t   minMaxMode            ;      // show act, min or max values
static boolean   graphMode             ;      // graphical output of servo signals
static int8_t    logMode               ;      // logging mode: -1 not available, 0 - disabled, 1 - enabled
static uint8_t   keyPressedTime;       ;      // time interval between key down and key up in 50ms steps
static uint16_t  lastReadTime          ;
static uint8_t   auxState              ;
static uint8_t   delayTime             ;      // x10 in milliseconds
static uint16_t  rawKey                ;
static uint8_t   vbat                  ;

// MWC data
static uint8_t   actMode[5]            ;      // activation modes for level, baro, compas, camStab, camTrig (6 bits, 0-2: Aux1 (Low, Mid, High), 3-5: Aux2 (Low, Mid, High)
static uint8_t   accMode, baroMode, magMode;
static uint8_t   accPres, baroPres, magPres;

static uint16_t  cycleTime             ; // this is the number in micro second to achieve a full loop, it can differ a little and is taken into account in the PID loop
static uint8_t   config                ; // Tri, Quad, Hex, etc.
static uint8_t   P8[4], I8[4], D8[4]   ; // 8 bits is much faster and the code is much shorter
static uint8_t   rcRate8               ; // dual rate
static uint8_t   rcExpo8               ; // expo
static uint8_t   accStrength           ; // AutoLevel Strength 0-250 (vers < v1.7pre)
static uint8_t   rollPitchRate         ; // reduces "P" and "D" of Roll/Pitch depending on the associated stick deviation
static uint8_t   yawRate               ; // reduces "P" and "D" of Yaw depending on Rudder stick deviation
static uint8_t   dynThrPID             ; // reduces "P" and "D" of Roll/Pitch when Throttle stick is above 50%

//static uint8_t activateCamStab8        ; // Camera Stabilisation (v1.7prebis)
//static uint8_t activateCamTrig8        ; // Camera Trigger       (v1.7prebis)

static int16_t   accSmooth   [3] MM_DIM;  // projection of smoothed and normalized gravitation force vector on x/y/z axis, as measured by accelerometer       
static int16_t   gyroData    [3] MM_DIM;  // from WMP+
static int16_t   altitude        MM_DIM;  // from barometer
static int16_t   heading         MM_DIM;  // from compas
static int16_t   angle       [2] MM_DIM;  // absolute angle inclination in deg (when Nunchuk/Acc present)
static int16_t   servo       [4] MM_DIM;  // depends on configuration
static int16_t   motor       [6] MM_DIM;  // depends on configuration 
static int16_t   rcHysteresis[8] MM_DIM;  // receiver input

// ----------------

#ifdef CHECK_INPUT
  static char       keySymbols [] = { '-', 'R', 'U', 'D', 'L', 'S', };
  PROGMEM prog_char jsStatusMsg[] = "\002Key\002Hor\002Vert";
#endif  

#ifdef JOYSTICK
  static int16_t jsHorCtr, jsVertCtr;
  static int16_t jsHorAct, jsVertAct;
    
  #ifdef NO_JS_SELECT_KEY
    #define INIT_JOYSTICK                       pinIn(JS_HOR_PIN); pinIn(JS_VERT_PIN); jsHorCtr = analogRead(JS_HOR_PIN); jsVertCtr = analogRead(JS_VERT_PIN);
  #else
    #define INIT_JOYSTICK pinIn(JS_SELECT_PIN); pinIn(JS_HOR_PIN); pinIn(JS_VERT_PIN); jsHorCtr = analogRead(JS_HOR_PIN); jsVertCtr = analogRead(JS_VERT_PIN);
  #endif
#else
  #define INIT_JOYSTICK
#endif

// ----------------

#ifdef APC220
  #define INIT_APC220               pinOut(APC220_ENABLE_PIN); enableAPC220(true);
  void enableAPC220(boolean value)  {digitalWrite(APC220_ENABLE_PIN, value);}
#else
  #define INIT_APC220
  #define enableAPC220(a)
#endif

// ----------------

#ifdef STATUS_LEDS
  #define INIT_LEDS                 pinOut(RED_LED); pinOut(YELLOW_LED); pinOut(GREEN_LED);

  #define red(a)                     digitalWrite(RED_LED,    a)
  #define yellow(a)                  digitalWrite(YELLOW_LED, a)
  #define green(a)                   digitalWrite(GREEN_LED,  a)
  void yellowOn ()                  {digitalWrite(YELLOW_LED, true); }
  void yellowOff()                  {digitalWrite(YELLOW_LED, false);}

  void blinkLed(uint8_t led, uint16_t del)                 {digitalWrite(led, HIGH); delay(del/2); digitalWrite(led, LOW); delay(del/2);}
  void blinkLed(uint8_t led, uint16_t del, uint8_t repeat) {while (repeat-- > 0) {blinkLed(led, del);}}
#else
  #define INIT_LEDS

  #define red(a)
  #define yellow(a)
  #define green(a)
  #define yellowOn()
  #define yellowOff()

  void blinkLed(uint8_t led, uint16_t del)                 {delay(del);}
  void blinkLed(uint8_t led, uint16_t del, uint8_t repeat) {delay(del * repeat);}
#endif // STATUS_LEDS

void blinkYellow500()                            {blinkLed(YELLOW_LED, 500);}
void blinkYellow4()                              {blinkLed(YELLOW_LED, 4);}

void blinkRed4()                                 {blinkLed(RED_LED, 4);}
void blinkRed(uint16_t del)                      {blinkLed(RED_LED, del);}
void blinkRed(uint16_t del, uint8_t repeat)      {blinkLed(RED_LED, del, repeat);}

// ----------------

void pinOut(uint8_t pin)                         {pinMode(pin, OUTPUT);}
void pinIn (uint8_t pin)                         {pinMode(pin, INPUT); digitalWrite(pin, HIGH);} // set pinMode and enable internal pullups

void delay50()                                   {delay(50);}
uint8_t getConfig()                              {return (config >= NUM_CONFIGS)? 0 : config;}

void initVars(uint8_t len, int16_t arr[] MM_DIM) {for (uint8_t i = 0; i < len; i++) {arr[i] MM_ACT = 1500;}}

// ----------------

void setup() {
  // make sure that the default chip select pin is set to output, even if you don't use it:
  pinOut(10);
  
  INIT_LCD
  INIT_BACKLIGHT
  INIT_JOYSTICK
  INIT_LEDS
  INIT_APC220
  INIT_TEST
  
  initVars(4, servo       );
  initVars(6, motor       );
  initVars(8, rcHysteresis);

  for (int8_t n = 0; n < 125; n++) {
    lcdPrintProg0(intro);
    delay(17);
    lcdClear();
  }

  if (!restoreParams()) {
    lcdPrintProg0(setMsgs[0]);
    enterValue(0);
  }

#ifdef LOGGING
  // see if the card is present and can be initialized:
  boolean succ = SD.begin(CHIP_SELECT);
  setBacklight();
  if (!succ) {
    lcdPrintProg1(noCard);
    logMode = LOG_ERROR;
    storeParams();
    blinkRed(500, 3);
  }
#endif

  serialBegin();

  state = 0;
  initialized = false;
  lcdPrintProg1(reading);
  uint8_t readBytes = serialRead();
  initialized = true;

  lcdPrintProg1(readMsg);
  format(6 + 16, readBytes, 2, 0);

  if (logMode <= NO_LOGGING) {
    refreshDisplay();
    blinkYellow500();
  }
#ifdef LOGGING
  else {
    openLog();
  }
#endif
}

/* ------ */

void loop() {
  
#ifdef LOGGING
  if (logMode > NO_LOGGING) {
    doLog();
    refreshDisplay();
  }
#endif  

  int8_t statusFlag = pgm_read_byte_near(nextState[state] + 5); 
  if (statusFlag != -1) {
    // update the current values from MultiWii board and refresh display
    uint16_t current = millis();
    if (current - lastReadTime > delayTime * 10) {
      lastReadTime = current;
      serialRead();
      showStatusLine();
    }
  }

  boolean v17prb = (vers >= V1_7_prb);
  
  int8_t key = getKeyPressed();
  if (key >=0) {
    int8_t next = getNextState(state, key);
    if (next >= 80) { // handle pseudo-states
      if (next < 85) {
        enterPid(next-80); 
      }
      else if (next < 87) {
        enterRates(next-85); 
      }
      else if (next == 87) {
        int8_t state0 = 0;
        while (state0 >= 0) {
          if (key >= 0) {
            lcdClear();
            lcdPrintProg0(setMsgs[state0]);
          }
          key = getKeyPressed();
          if      (key == KEY_LEFT) { state0 = -1; }
          else if (key == KEY_UP  ) { state0 = (state0 + NUM_SETTINGS - 1) % NUM_SETTINGS; }
          else if (key == KEY_DOWN) { state0 = (state0                + 1) % NUM_SETTINGS; }
          else if (key >= 0)        { enterValue(state0); setBacklight(); }
        }
      } 
      else if (next == 88) {
        int8_t state0 = 0;
        while (state0 >= 0) {
          if (key >= 0) {
            lcdClear();
            lcdPrintProg0(profStates[state0]);
            lcdPrint(" Profile");
          }
          key = getKeyPressed();
          if      (key == KEY_LEFT) { state0 = -1; }
          else if (key == KEY_UP  ) { state0 = (state0 + 3) % 4; }
          else if (key == KEY_DOWN) { state0 = (state0 + 1) % 4; }
          else if (key >= 0) {
            if (state0 == 3) {
              extractProfile(defaultProfile);
            }
            else {
              lcdWrite(':'); 
              while (true) {
                if (key >= 0) {
                  profileMsg[2] = ('0' + profile/10);
                  profileMsg[3] = ('0' + profile%10);
                  getProfileName(profileMsg+6);
                  
                  lcdPrint(16, profileMsg);
                }
                key = getKeyPressed();
                if      (key == KEY_LEFT) { break; }
                else if (key == KEY_UP  ) { profile = (profile+15) % 16; }
                else if (key == KEY_DOWN) { profile = (profile+ 1) % 16; }
                else if (key >= 0) {
                  if      (state0 == 0) { restoreProfile(); } // load profile
                  else if (state0 == 1) { storeProfile();   } // save profile
                  else                  { deleteProfile();  } // clear profile
                  break;
                }
              }
            }
          }
        }
      } 
      else if (next == 89) {
#ifdef MINMAX
  #ifdef GRAPHIC
        if (key == KEY_SELECT || keyPressedTime > 10) {
          graphMode = !graphMode;
        }
        else
  #endif
        {
          minMaxMode = (minMaxMode + 1) % 3; 
          showMinMax();
        }
#else        
  #ifdef GRAPHIC
        graphMode = !graphMode;
  #endif
#endif
      }
      else if (next == 90) {
        enterAuxState();
      }
      else if (next == 91) {
        int8_t offs = (key == KEY_UP)? 4 : 1;
        auxState = (auxState + offs) % 5;
      }
#ifdef LOGGING
      else if (next == 95) {
        int8_t state0 = 0;
        while (state0 >= 0) {
          if (key >= 0) {
            lcdClear();
            lcdPrintProg0(logStates[state0]);
          }
          key = getKeyPressed();
          if      (key == KEY_UP)   {state0 = (state0 + 2) % 3;}
          else if (key == KEY_DOWN) {state0 = (state0 + 1) % 3;}
          else if (key == KEY_LEFT) {state0 = -1;}
          else if (key >= 0) {
            if (state0 == 0) {
              // start logging
              if (logMode >= NO_LOGGING) {
#ifdef LOGVIEW
                logMode = confirm(formatMsg)? LF_LOGVIEW : LF_BINARY;
#else
                logMode = LF_BINARY;
#endif

                if (confirm(haltReqMsg)) {
                  storeParams();
                  setError(logEnbMsg);
                }
                openLog();
                state0 = -1;
              }
              else {
                blinkRed(250, 2);
              }
            }
            else if (state0 == 1) {
              playbackLog(); 
            }
            else {
              // delete log file
              int fileNo = 999;
              while (fileNo >= 0) {
                fileNo = selectLogfile(fileNo, false);
                if (fileNo >= 0) {
                  boolean succ = SD.remove(fileName);
                  setBacklight();
                  
                  if (succ) {
                    blinkYellow500();
                  }
                  else {
                    lcdPrintProg1(deleteFailed);
                    blinkRed(250, 6);
                    lcdPrintProg1(emptyLine);
                  }
                }
              }
            }
          }
        }
      }
#endif
      else if (next == 99) { 
#ifdef TEST
        if (key == KEY_SELECT) {
          config++;
          if (config >= NUM_CONFIGS) {
            config = 1;
          }
          lcdPrintProg1(configIds + configOffsets[getConfig()]);
          blinkYellow500();
        }
#endif          
        state = getConfig();         // select state depending on config
        if (state > 0) { 
          state += CONF_OFFSET; 
        } 
        else {
          state = 6;
        }
      }
      else if (next == 100) {
#ifdef CHECK_INPUT
        if (keyPressedTime > 10) {
          lcdClear();
          lcdPrintProg0(jsStatusMsg);

          while (true) {
            int8_t key = getKey();
            
            lcdWrite(16, keySymbols[key+1]);
            format      ( 2 + 16, rawKey);
  #ifdef JOYSTICK            
            formatSigned( 7 + 16, jsHorAct  - jsHorCtr);
            formatSigned(12 + 16, jsVertAct - jsVertCtr);
  #endif
            key++;
            red   (key & 0x1);
            yellow(key & 0x2);
            green (key & 0x4);

            delay50();
          }
        }
        else
#endif
        if (v17prb && confirm(calibrateMsg)) {
          Serial.write('D');
          state = 4; 
        }
      } 
      else if (next == 101) {
        if (v17prb) {
          state = 13;
        }
        else if (key == KEY_UP) {
          state = 12;
        }
        else {
          state = 9;
        }
      }
      else if (next == 102) {
        if (v17prb) {
#ifdef BATTERY_LEVEL
          state = 34;
#else
          state = 20;
#endif          
        }
        else {
          state = 0;
        }
      }

      next = state;
    }
    if (next >= 0 && logMode <= NO_LOGGING) {
      state = next;
      refreshDisplay();
    }
  }
}

void refreshDisplay() {
  lcdClear();

  // two levels of indirection...
  int8_t           index = pgm_read_byte_near(nextState[state] + 4);            // this is the index in the msgs array
  prog_char const* cptr0 = (prog_char const*) pgm_read_word_near(msgs + index); // this is the adress of the msg
                   index = pgm_read_byte_near(nextState[state] + 5); 
  prog_char const* cptr1 = (index >= 0)? (prog_char const*) pgm_read_word_near(msgs + index) : emptyLine;
  lcdPrintProg2(cptr0, cptr1);

  showStatusLine();
}

void showStatusLine() {
  uint8_t v5 = (vers < V1_6)? 5 : 4; // covers the "swapped escs for Y6/hex" bug

  if (state == 0) {
    lcdPrintProg1(configIds + configOffsets[getConfig()]);
    lcdWrite(' ');
    int8_t index = (logMode == OFFLINE)? 2 : accMode;
    lcdPrintProg(8 + 16, modeMsg[index]);
  }
  else if (state == 1) {
    if (keyPressedTime > 10) {
      lcdPrintProg1(versions[vers]);
      lcdWriteProg(metaData);
      lcdWrite(baudRate + '0');
      enterKey();
      lcdPrintProg1(credits);
      enterKey();
    }
    if (accPres > 1) {
      statusMsg[0] = 'A';
      statusMsg[2] = 'c';
    }
    else {
      statusMsg[0] = 'N';
      statusMsg[2] = 'k';
    }
    statusMsg[ 3] = (accPres )? (accMode ? '1' : '0') : '-';
    statusMsg[ 9] = (baroPres)? (baroMode? '1' : '0') : '-';
    statusMsg[15] = (magPres )? (magMode ? '1' : '0') : '-';
    
    lcdPrint(16, statusMsg);
  }
  else if (state >= 9 && state <= 13) {
    int8_t idx = state - 10;
    if (idx < 0) {idx = 0;}

    format( 1 + 16, P8[idx], 4, 2);    
    format( 6 + 16, I8[idx], 5, 4);
    if (state < 13) {
      formatSigned(12 + 16, -D8[idx], 3, 0);
    }
  }
  else if (state == 14) { // Dual Rate, Expo, AutoLevel Strength
    format( 0 + 16, rcRate8 * 2);
    format( 6 + 16, rcExpo8    );
    if (vers < V1_7_prb) {
      format(12 + 16, accStrength);
    }
  }
  else if (state == 15) { // R/P Rate, Yaw Rate, Throttle PID
    format( 0 + 16, rollPitchRate);
    format( 6 + 16, yawRate      );
    format(12 + 16, dynThrPID    );
  }
  else if (state == 16) {  // Gyro
    for (uint8_t i = 0; i < 3; i++) { formatSigned(i*6 + 16, gyroData[i] MM_VAL); }
  }
  else if (state == 17) { // Acc
    for (uint8_t i = 0; i < 3; i++) { formatSigned(i*6 + 16, accSmooth[i] MM_VAL); }
  }
  else if (state == 18) { // Alt, Head, Cycletime
    formatSigned( 0 + 16, altitude MM_VAL);
    formatSigned( 6 + 16, heading  MM_VAL);
    format(12 + 16, cycleTime);
  }
  else if (state == 19) { // Angles
    if (!graphMode) {
      formatSigned( 1 + 16, angle[0] MM_VAL);
      formatSigned(10 + 16, angle[1] MM_VAL);
    }
#ifdef GRAPHIC
    else {
      drawAngle(0);
      drawAngle(1);
    }
#endif
  }
  else if (state == 21) { // Roll, Pitch, Yaw
    for (uint8_t i = 0; i < 3; i++) {formatScaledSym( i*6 + 16, rcHysteresis[i] MM_VAL);}
  }
  else if (state == 22) { // Throttle, Aux1, Aux2
    for (uint8_t i = 0; i < 3; i++) {formatScaledSym( i*6 + 16, rcHysteresis[i+3] MM_VAL);}
  }
  else if (state == 23) { // Cam_Roll, Cam_Pitch
    formatScaledSym( 6 + 16, rcHysteresis[CAM_ROLL] MM_VAL);
    formatScaledSym(12 + 16, rcHysteresis[CAM_PIT ] MM_VAL);
  }
  else if (state == 24) { // Tri: Left, Right, Rudder, Rear
    formatScaledAbs( 3, motor[2] MM_VAL);
    formatScaledAbs(12, motor[1] MM_VAL);

    formatScaledSym( 3 + 16, servo[0] MM_VAL);
    formatScaledAbs(12 + 16, motor[0] MM_VAL);
  }
  else if (state == 25) { // Quad+: Front, Rear, Left, Right
    formatScaledAbs( 3, motor[3] MM_VAL);
    formatScaledAbs(12, motor[0] MM_VAL);

    formatScaledAbs( 3 + 16, motor[2] MM_VAL);
    formatScaledAbs(12 + 16, motor[1] MM_VAL);
  }
  else if (state == 26) { // QuadX: Front Left, Front Right, Rear Left, Rear Right
    formatScaledAbs( 3, motor[3] MM_VAL);
    formatScaledAbs(12, motor[1] MM_VAL);

    formatScaledAbs( 3 + 16, motor[2] MM_VAL);
    formatScaledAbs(12 + 16, motor[0] MM_VAL);
  }
  else if (state == 27) { // Bi: Motor 0, Motor 1, Servo 0, Servo 1 
    formatScaledAbs( 3, motor[0] MM_VAL);
    formatScaledAbs(12, motor[1] MM_VAL);

    formatScaledSym( 3 + 16, servo[0] MM_VAL);
    formatScaledSym(12 + 16, servo[1] MM_VAL);
  }
  else if (state == 28) { // Gimbal: Roll, Pitch
    formatScaledSym( 7, servo[1] MM_VAL);
    formatScaledSym( 7 + 16, servo[0] MM_VAL);
  }
  else if (state == 29) { // Y6: Up (Left, Mid/Rear, Right), Down (Left, Mid, Right)
    formatScaledAbs( 2, motor[2   ] MM_VAL);
    formatScaledAbs( 7, motor[0   ] MM_VAL);
    formatScaledAbs(12, motor[1   ] MM_VAL);

    formatScaledAbs( 7 + 16, motor[3   ] MM_VAL);
    formatScaledAbs( 2 + 16, motor[9-v5] MM_VAL);
    formatScaledAbs(12 + 16, motor[v5  ] MM_VAL);
  }
  else if (state == 30) { // Hex: Front (Left, Mid, Right), Rear (Left, Mid, Right)
    formatScaledAbs( 2, motor[3   ] MM_VAL);
    formatScaledAbs( 7, motor[v5  ] MM_VAL);
    formatScaledAbs(12, motor[1   ] MM_VAL);

    formatScaledAbs( 2 + 16, motor[2   ] MM_VAL);
    formatScaledAbs( 7 + 16, motor[9-v5] MM_VAL);
    formatScaledAbs(12 + 16, motor[0   ] MM_VAL);
  }
  else if (state == 31) { // Flying Wing: Servo 0-3 
    formatScaledSym( 3, servo[0] MM_VAL);
    formatScaledSym(12, servo[1] MM_VAL);

    formatScaledSym( 3 + 16, servo[2] MM_VAL);
    formatScaledSym(12 + 16, servo[3] MM_VAL);
  }
  else if (state == 32) { // Y4: Motor 0-3 
    formatScaledSym( 3, motor[3] MM_VAL);      // front left, right
    formatScaledSym(12, motor[1] MM_VAL);

    formatScaledSym( 3 + 16, motor[0] MM_VAL);  // rear 1+2
    formatScaledSym(12 + 16, motor[2] MM_VAL);
  }
  else if (state == 33) { // HexX: Left (Front, Mid, Rear), Right (Front, Mid, Rear)
    formatScaledAbs( 2, motor[3] MM_VAL);
    formatScaledAbs( 7, motor[5] MM_VAL);
    formatScaledAbs(12, motor[2] MM_VAL);

    formatScaledAbs( 2 + 16, motor[1] MM_VAL);
    formatScaledAbs( 7 + 16, motor[4] MM_VAL);
    formatScaledAbs(12 + 16, motor[0] MM_VAL);
  }
  else if (state == 20) { // Activation Mode
    lcdPrintProg1(auxStates[auxState]);
    printAuxState(actMode[auxState]);
  }
#ifdef BATTERY_LEVEL
  else if (state == 34) { // Battery Level
    lcdPrintProg1(batLevelMsg);
    format(25, vbat, 4, 2);

    if (vbat > 0) {
      green (vbat >  VBATLEVEL1);      
      yellow(vbat >  VBATLEVEL3 && vbat <= VBATLEVEL0);
      red   (vbat <= VBATLEVEL2);      
    }
  }
#endif  
}

void printAuxState(int8_t mode) {
  for (int8_t i = 0; i < 6; i++) {
    int8_t col = i + 7;
    if (i > 2) {col += 3;}
    lcdWrite(col + 16, (mode & auxMasks[i])? '*' : 165);
  }
}

int8_t getNextState(int8_t state, int8_t key) { 
  int8_t col = (key == KEY_SELECT)? KEY_RIGHT : (key == STATUS)? STATUS-1 : key;
#ifdef MINMAX
  if (state >= 4 && state <= 6 && col == KEY_RIGHT) {showMinMax();}
#endif  

#ifdef BATTERY_LEVEL
  if (state == 34) {
    green(false); yellow(false); red(false);
  }
#endif

  int8_t next = pgm_read_byte_near(nextState[state] + col);

#ifdef LOGGING
  // disable "Logging" state when initialisation of SD Card failed 
  if (next == LOG_STATE && logMode == LOG_ERROR) {next = (key == KEY_UP)? 8 : 0;}
#endif

  return next;
}

// ----------------

void lcdClear()                                                      {lcd.clear();}
void lcdBlink()                                                      {lcd.blink();  yellowOn();}
void lcdCursor()                                                     {lcd.cursor(); yellowOn();}
void lcdCreateChar(int8_t idx, uint8_t* array)                       {lcd.createChar(idx, array);}

void lcdWrite(int8_t digit)                                          {lcd.write(digit);}
void lcdWrite(uint8_t pos, int8_t digit)                             {lcdSetCursor(pos); lcdWrite(digit);}

void lcdPrint(char const *msg)                                       {lcd.print(msg);}
void lcdPrint(uint8_t pos, char const msg[])                         {lcdSetCursor(pos); lcdPrint(msg);}

void lcdPrintProg0(prog_char const line[])                           {lcdPrintProg( 0, line);}
void lcdPrintProg1(prog_char const line[])                           {lcdPrintProg(16, line);}
void lcdPrintProg2(prog_char const line0[], prog_char const line1[]) {lcdPrintProg0(line0); lcdPrintProg1(line1);}
void lcdPrintProg (uint8_t pos, prog_char const line[])              {lcdSetCursor(pos); lcdWriteProg(line);}

void lcdWriteProg(prog_char const msg[]) {
  char* lptr = lineBuffer;
  char  last = -1;
  for (int8_t i = 0, k = 0; last != '\0' && k < 16; i++) {
    last = pgm_read_byte_near(msg + i);
    if (last <= 16) {
      // a char ch < 16 represent empty strings of length ch - this saves a lot of memory
      for (int8_t j = 0; j < last && k < 16; j++) {*lptr++ = ' '; k++;}
    }
    else {
      *lptr++ = last;
      k++;
    }
  }
  *lptr++ = '\0';

  lcdPrint(lineBuffer);
}

void lcdSetCursor(uint8_t pos) {
  uint8_t col = pos % 16;
  uint8_t row = pos / 16;
  lcd.setCursor(col, row);
}

void setError(prog_char const msg[]) {
  lcdPrintProg2(msg, haltMsg);
  while (true) {
    blinkRed(2000);
  }
}

// ----------------

void enterPid(int8_t index) {
  uint8_t col[] = {5, 11, 15, };
  int val[3], minv[3] = { 0, 0, -40 }, maxv[3] = { 250, 250, 0 };

  int8_t idx = index - 1;
  if (idx < 0) {idx = 0;}
 
  val[0] =  P8[idx]; 
  val[1] =  I8[idx]; 
  val[2] = -D8[idx];
  
  int8_t local    =  0; // local state (-1..4)
  int8_t lastKey  = -1;
  int    count    =  0;
  int    repeat   =  0;
  int8_t maxState =  3;
  if (index == 4) {maxState = 2;}
  
  lcdBlink();
  while (local >= 0 && local < maxState) {
    lcdSetCursor(col[local] + 16);

    int8_t key = getKey();
    if (key < 0) {
      lastKey = key;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      if (key == KEY_LEFT) {
        local--;
      }
      else if (key == KEY_SELECT || (key == KEY_RIGHT && local == 2)) {
        if (index == 0) {
          P8[1] =  val[0];
          I8[1] =  val[1];
          D8[1] = -val[2];
        }
        P8[idx] =  val[0];
        I8[idx] =  val[1];
        D8[idx] = -val[2];
        
        serialWrite();
        local = -1;
      }
      else if (key == KEY_RIGHT) {
        local++; 
      }
      else {
        if (repeat > 3 || (count % 8) == 0) {
          int tmp = val[local];
          if (key == KEY_UP) {
            tmp = max(tmp-1, minv[local]);
          }
          else {
            tmp = min(tmp+1, maxv[local]);
          }
          val[local] = tmp;
          if      (local == 0) { format      ( 1 + 16, tmp, 4, 2); }
          else if (local == 1) { format      ( 6 + 16, tmp, 5, 4); }    
          else if (local == 2) { formatSigned(12 + 16, tmp, 3, 0); } 
          repeat++;
        }
      }
    }
    count++;
    delay50();
  }
  waitForKeyUp();
}

void enterRates(int8_t index) {
  uint8_t val[3];
 
  if (index == 0) {
    val[0] = rcRate8;
    val[1] = rcExpo8;
    val[2] = accStrength;
  }
  else {
    val[0] = rollPitchRate;
    val[1] = yawRate;
    val[2] = dynThrPID;
  }
  
  int8_t local    =  0;  // local state (-1..4)
  int8_t lastKey  = -1;
  int    count    =  0;
  int    repeat   =  0;
  int8_t maxState = (index == 0 && vers >= V1_7_prb)? 2 : 3;

  lcdBlink();
  while (local >= 0 && local < maxState) {
    uint8_t col = local * 6 + 4;
    if (col > 15) {col = 15;}
    lcdSetCursor(col + 16);
    
    int8_t key = getKey();
    if (key < 0) {
      lastKey = -1;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      if (key == KEY_LEFT) {
        local--;
      }
      else if (key == KEY_SELECT || (key == KEY_RIGHT && local == 2)) {
        if (index == 0) {
          rcRate8       = val[0];
          rcExpo8       = val[1];
          accStrength   = val[2];
        }
        else {
          rollPitchRate = val[0];
          yawRate       = val[1];
          dynThrPID     = val[2];
        }

        serialWrite();
        local = -1;
      }
      else if (key == KEY_RIGHT) {
        local++; 
      }
      else {
        if (repeat > 3 || (count % 8) == 0) {
          int tmp = val[local];
          if (key == KEY_UP) {
//            uint8_t minVal = (index == 0 && local == 0)? 10 : 0;
            tmp--;
            if (tmp < 0) {
              tmp = 0;
            }
          }
          else {
            uint8_t maxVal = (index == 0 && local != 1)? 250 : 100;
            tmp++;
            if (tmp > maxVal) {
              tmp = maxVal;
            }
          }
          val[local] = tmp;
          if (index == 0 && local == 0) {
            tmp *= 2;
          }
          format(local * 6 + 16, tmp);
          repeat++;
        }
      }
    }
    count++;
    delay50();
  }
  waitForKeyUp();
}

boolean enterName(char name[]) {
  int8_t local   =  0; // local state (-1..8)
  int8_t lastKey = -1;
  int    count   =  0;
  int    repeat  =  0;
  int8_t maxCh   =  sizeof(charsInName)-2;
  
  for (int8_t i = 0; i < 8; i++) {
    if (name[i] == '-') {
      name[i] = ' ';
    }
  }
  
  int8_t idx = findChar(name[local]);
  
  lcdCursor();
  lcdWrite(6+16, name[0]);
  lcdSetCursor(6 + 16);
  
  while (local >= 0 && local < 8) {
    int8_t key = getKey();
    if (key < 0) {
      lastKey = -1;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      boolean refresh = false;
      if (key == KEY_LEFT) {
        local--;
        refresh = (local >= 0);
        if (refresh) {
          idx = findChar(name[local]);
        }
      }
      else if (key == KEY_RIGHT) { 
        local++; 
        refresh = (local < 8);
        if (refresh) { // handle RIGHT in third column as SELECT
          idx = findChar(name[local]);
        }
      }
      else if (key == KEY_SELECT) {
        local = 8;
      }
      else {
        if (repeat > 3 || (count % 8) == 0) {
          if (key == KEY_DOWN && idx < maxCh) {
            idx++;
            refresh = true;
          }
          else if (key == KEY_UP && idx > 0) {
            idx--;
            refresh = true;
          }
          if (refresh) {
            name[local] = charsInName[idx];
          }
        }
      }

      if (refresh) {
        lcdWrite    (local + 6 + 16, name[local]);
        lcdSetCursor(local + 6 + 16);
        repeat++;
      }
    }
    count++;
    delay50();
  }
  waitForKeyUp();
  
  return (local == 8);
}

int8_t findChar(char ch) {
  int8_t res = 0;
  int8_t len = sizeof(charsInName)-1;
  for (int8_t i = 1; res < 1 && i < len; i++) {
    if (charsInName[i] == ch) {
      res = i;
    }
  }
  return res;
}

/* ----- */

void enterAuxState() {
  lcdCursor();
  
  int8_t local = 0; // local state (-1..6)
  int8_t mode  = actMode[auxState];
  int8_t key   = 0;

  while (local >= 0 && local < 6) {
    if (key >= 0) {
      printAuxState(mode);
      int8_t col = local + 7 + 3 * (local > 2);
      lcdSetCursor(col + 16);
    }
    key = getKeyPressed();
    if      (key == KEY_LEFT)   {local--;}
    else if (key == KEY_RIGHT)  {local++;}
    else if (key == KEY_SELECT) {local = 6;}
    else if (key >= 0)          {mode ^= auxMasks[local];}

    delay50();
  }
  waitForKeyUp();
  
  if (local == 6) {
    actMode[auxState] = mode;
    serialWrite();    
  }
}

/* ----- */
    
void enterValue(uint8_t index) {
  lcdWrite(':');
  yellowOn();
  
  int val = backlight;
  if      (index == 0) { val = vers;      }
  else if (index == 1) { val = delayTime; lcdPrintProg1(delaySuff); }
  else if (index == 2) { val = baudRate;  format(7 + 16, 0); }
//  else                 { val = backlight; }

  int    repeat  = 0;
  int    count   = 0;
  int8_t lastKey = -1;

  while (true) {
    if      (index == 0) { lcdPrintProg(5 + 16, versions[val]); }
    else if (index == 2) { format(5 + 16, baudRates[val]*12, 4, 0); }
    else                 { format(5 + 16, val, 3, 0);             }
    
    int8_t key = getKey();
    if (key < 0) {
      lastKey = -1;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      if (key == KEY_UP || key == KEY_DOWN) {
        if (repeat > 3 || (count % 8) == 0) {
          int8_t  incr = 1;
          uint8_t maxV = 250;

          if      (index == 0) { maxV = sizeof(frameLength)-1; }
          else if (index == 2) { maxV = 8; }
          else                 { incr = 5; }
          if (key == KEY_UP)   { incr = -incr; }

          val = constrain(val+incr, 0, maxV);

          if (index == 3) {
            analogWrite(BACKLIGHT_PIN, val);
          }

          repeat++;
        }
      }
      else {
        if (key != KEY_LEFT) {
          if      (index == 0) { vers      = val; }
          else if (index == 1) { delayTime = val; }
          else if (index == 2) { baudRate  = val; }
          else                 { backlight = val; }
  
          storeParams();
        }
        waitForKeyUp();
        return; 
      }
    }
    count++;
    delay50();
  }
}

// ----------------

void format      (uint8_t pos, int val)                          {format      (pos, val, 4, 0);}
void format      (uint8_t pos, int val, int8_t len, int8_t prec) {formatNum   (pos, val, len, prec, false);}

void formatSigned(uint8_t pos, int val)                          {formatSigned(pos, val, 4, 0);}
void formatSigned(uint8_t pos, int val, int8_t len, int8_t prec) {formatNum   (pos, val, len, prec, true);}

void formatNum   (uint8_t pos, int val, int8_t len, int8_t prec, bool sgn) {
  char temp[len+1];
  int8_t low = 0;
  if (sgn) {
    temp[0] = (val > 0)? '+' : (val < 0)? '-' : ' ';
    low++;
  }
  val = abs(val);
  for (int8_t i = len-1; i >= low; i--) {
    if (len-i == prec) {
      temp[i] = '.';
    }
    else {
      temp[i] = '0' + (val % 10);
      val    /= 10;
    }
  }
  if (val > 0) { // overflow
    for (int8_t i = len-1; i >= 0; i--) {
      temp[i] = '#';
    }
  }
  temp[len] = '\0';

  lcdPrint(pos, temp);
}

void formatScaledSym(uint8_t pos, int val) { // -100..+100
  val = (val - 1500)/5;
#ifdef GRAPHIC
  if (graphMode) {
    formatGraphic(pos, (val+100)/2);
  }
  else
#endif
  {
    formatSigned(pos, val);
  }
}

void formatScaledAbs(uint8_t pos, int val) { // 0..100
  val = (val - 1000)/10;
  if (!graphMode) {
    formatSigned(pos, val);
  }
#ifdef GRAPHIC
  else {
    formatGraphic(pos, val);
  }
#endif
}

#ifdef GRAPHIC
static boolean charsInitialized;    // chars for servo signals are initialized

void formatGraphic(uint8_t pos, int val) { // 0..100
  if (!charsInitialized) {
    charsInitialized = true;
  
    static byte bars[8] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x15, };
    static byte less[8] = { 0x00, 0x04, 0x0C, 0x1C, 0x0C, 0x04, 0x00, 0x15, };
    static byte grt [8] = { 0x00, 0x04, 0x06, 0x07, 0x06, 0x04, 0x00, 0x15, };

    byte pattern = 0x10;
    for (int8_t i = 0; i <= 5; i++) {
      for (int8_t j = 0; j < 7; j++) {
        bars[j] = pattern;
      }
      lcdCreateChar(i, bars);
      pattern >>= 1;
    }
    lcdCreateChar(6, less);
    lcdCreateChar(7, grt);
  }  
  
  char bar[] = { 5, 5, 5, 5 };
  
  if      (val < -100 || val > 200) { bar[0] = 6; bar[3] = 7; }    // invalid
  else if (val <    0)              { bar[0] = 6; }                // <...
  else if (val >= 100)              { bar[3] = 7; }                // ...>
  else                              { bar[val/25] = (val%25)/5; }  // ..|.

  lcdSetCursor(pos);
  for (int8_t i = 0; i < 4; i++) {
    lcdWrite(bar[i]); 
  }
}

/* ----- */

static uint8_t angleChars[4][8];

int8_t round8(float val) {return (int8_t) (val + 0.5f);}

void drawAngle(int8_t index) {

  // clear char arrays
  uint8_t *cptr = angleChars[0];
  for (int8_t i = 0; i < 32; i++) {
    *cptr++ = 0;
  }

  int8_t  charOffset = index * 4;
  int16_t alpha      = - angle[index] MM_VAL;
  if (alpha >= -360 && alpha <= 360) {
    float ang  = radians(alpha);
    float sinA = sin(ang);
    float cosA = cos(ang);
    
  //  if (abs(sinA) > 0.25f) {
  //    angleChars[0][4] = 0x14;  // mark horizon
  //    angleChars[3][4] = 0x05;
  //  }
    
//    int8_t x0 = (int8_t) (- cosA * 10.0f + 0.5f);
//    int8_t y0 = (int8_t) (- sinA * 10.0f + 0.5f);
    int8_t x0 = round8(- cosA * 10.0f);
    int8_t y0 = round8(- sinA * 10.0f);
    
    drawLine(x0, y0, -x0, -y0);
    
//    int8_t x1 = (int8_t) (- sinA * 3.0f + 0.5f);
//    int8_t y1 = (int8_t) (  cosA * 3.0f + 0.5f);
    int8_t x1 = round8(- sinA * 3.0f);
    int8_t y1 = round8(  cosA * 3.0f);
  
    drawLine(0, 0, x1, y1);
  }
  else {
    angleChars[1][4] = 0x06;  
    angleChars[2][4] = 0x18;  
  }
  
  for (int8_t i = 0; i < 4; i++) {
    lcdCreateChar(charOffset + i, angleChars[i]);
    lcdWrite(16 + 2 + index * 8 + i, charOffset + i);
  }
  
  charsInitialized = false;  
}

void drawLine(int8_t x0, int8_t y0, int8_t x1, int8_t y1) {
  // Bresenham again...
  
  int8_t dx =  abs(x1-x0), sx = x0<x1 ? 1 : -1;
  int8_t dy = -abs(y1-y0), sy = y0<y1 ? 1 : -1; 
  int8_t err = dx+dy, e2; /* error value e_xy */
 
  for(;;){  /* loop */
    setPixel(x0, y0);
    if (x0==x1 && y0==y1) break;
    e2 = 2*err;
    if (e2 >= dy) { err += dy; x0 += sx; } /* e_xy+e_x > 0 */
    if (e2 <= dx) { err += dx; y0 += sy; } /* e_xy+e_y < 0 */
  }  
}

void setPixel(int8_t x, int8_t y) { 
  // x: -10..+9, y: -3..+4
  
  if (x >= -10 && x < 10 && y >= -3 && y < 5) {
    int8_t nrow = 4 - y;   // 0..7
    int8_t ncol = x + 10;  // 0..19
    int8_t idx  = ncol / 5;
    int8_t pcol = ncol % 5;
    
    angleChars[idx][nrow] |= (0x10 >> pcol);
  }
}
#endif // GRAPHIC

// ----------------

#ifdef MINMAX
void showMinMax() {
  lcdClear();
  lcdPrintProg0(minMaxDsc[minMaxMode]);
  blinkYellow500();
}
#endif // MINMAX

// ----------------

boolean confirm(prog_char const msg[]) {
  lcdPrintProg2(msg, yesNoMsg);
  waitForKeyUp();
  yellowOn();
    
  int8_t key = -1;
  while (key != KEY_LEFT && key != KEY_RIGHT) {
    key = getKeyPressed();
  }      
  waitForKeyUp();
  return (key == KEY_RIGHT);
}

int8_t enterKey() {
  yellowOn();
  int8_t key = -1;
  while ((key = getKeyPressed()) < 0);
  yellowOff();
  return key;
}

int8_t getKeyPressed() {
  uint8_t cnt = 0;
  int8_t  key = getKey();
  if (key >= 0) {
    while (rawGetKey() >= 0) {
      if (cnt < 255) {
        cnt++;
      }
      if (cnt > 20) {
        blinkRed4();
      }
      delay(50);
    }
  }
  keyPressedTime = cnt;
  return key;
} 

void waitForKeyUp() {
  while (rawGetKey() >= 0);

  yellowOff();
  lcd.noCursor();
  lcd.noBlink();
}

int8_t getKey() {
  static int8_t oldKey = -1;

  int8_t key = rawGetKey();
  if (key != oldKey) {
    delay50();		// wait for debounce time
    key = rawGetKey();
    if (key != oldKey) {			
      oldKey = key;
    }
  }
  return key;
}

#ifndef HWKEYS
int adc_key_val[NUM_KEYS] = { 30, 175, 360, 535, 760 }; // thresholds of the cascading resistors
#endif

int8_t rawGetKey() {
#ifdef HWKEYS
  if      (digitalRead(HW_RIGHT) ==0) {return KEY_RIGHT; }
  else if (digitalRead(HW_UP)    ==0) {return KEY_UP;    }
  else if (digitalRead(HW_DOWN)  ==0) {return KEY_DOWN;  }
  else if (digitalRead(HW_LEFT)  ==0) {return KEY_LEFT;  }
  else if (digitalRead(HW_SELECT)==0) {return KEY_SELECT;}
#else
  rawKey = analogRead(KEY_PIN);   // read the value from the sensor  
  for (uint8_t k = 0; k < NUM_KEYS; k++) {
    if (rawKey < adc_key_val[k]) {
      return k;
    }
  }
#endif

#ifdef JOYSTICK
 #ifndef NO_JS_SELECT_KEY
  int sel  = analogRead(JS_SELECT_PIN);
 #else
  int sel = 1023;    
 #endif
  jsHorAct  = analogRead(JS_HOR_PIN);
  jsVertAct = analogRead(JS_VERT_PIN);

  if (sel < 1000 || jsHorAct < 1000 || jsVertAct < 1000) { // is joystick present? (int. pullups are on)
 #ifdef JS_INV_SELECT_KEY
    boolean select = (sel < 4);    // LOW when pressed 
 #else
    boolean select = (sel > 1020); // HIGH when pressed
 #endif
 #ifndef NO_JS_SELECT_KEY
    if (select) {
      return KEY_SELECT;
    }
    else
 #endif
    {
      int hor  = (jsHorAct  - jsHorCtr ) * JS_HOR_SIGN;
      int vert = (jsVertAct - jsVertCtr) * JS_VERT_SIGN;
  
      if (abs(vert) < JS_DEADBAND) {
        if      (hor < -5 * JS_DEADBAND)  {return KEY_RIGHT;}
        else if (hor >  5 * JS_DEADBAND)  {return KEY_LEFT; }
      }
      else if (abs(hor) < JS_DEADBAND) {
        if      (vert < -5 * JS_DEADBAND) {return KEY_UP;   }
        else if (vert >  5 * JS_DEADBAND) {return KEY_DOWN; }
      }
    }
  }
#endif

  return -1;
}

// ----------------

static uint8_t* bptr;
static uint8_t  buffer[100];      // max. number of bytes read from serial

boolean serialAvailable()         {return Serial.available();}
void    serialBegin()             {Serial.begin(long(baudRates[baudRate])*1200);}  
uint8_t readByte()                {Serial.read();}  
//void serialWriteBuffer()          {Serial.write(buffer, bptr - buffer); Serial.flush();}

void serialize8   ( int8_t a)     {*bptr++ = a;}
void serialize16  (int16_t a)     {*bptr++ = a; *bptr++ = a/256;}

void deserialize8 (  int8_t& val) {val = *bptr++;}
void deserialize8 ( uint8_t& val) {val = *bptr++;}
void deserialize16( int16_t& val) {uint8_t low = *bptr++; uint8_t hi = *bptr++; val = (int16_t)  (hi * 256 + low);}
void deserialize16(uint16_t& val) {uint8_t low = *bptr++; uint8_t hi = *bptr++; val = (uint16_t) (hi * 256 + low);}

void deserialize16(int16_t arr[]) {
  deserialize16(arr[0]); 
  if (initialized) {
    arr[1] = min(arr[0], arr[1]); 
    arr[2] = max(arr[0], arr[2]);
  }
  else {
    arr[1] = arr[2] = arr[0];
  }
}

void setOffline(prog_char const* msg) {
  lcdPrintProg1(msg);
  blinkRed(250, 6);
  logMode = OFFLINE;
  storeParams();
  refreshDisplay();
}

#ifndef TEST
uint8_t serialRead() {
  
  if (logMode != OFFLINE) {
    Serial.flush();
  
    boolean refresh = false;
    uint8_t len    = frameLength[vers];
    uint8_t fcnt   = 0;
    int     retry  = 0;
    int     avail  = 0;
    char    tag    = 'A';
    
    if (state == 34) {
      tag = '0';    // OSD
      len = 50;
    }
    
    while ((avail = serialAvailable()) <= len && retry < 600) {
      if (!initialized && getKey() == KEY_LEFT) {
        // cancel loop
        setOffline(cancelRead);
        return 0;
      }
      if ((retry % 20) == 0) {
        if (avail > 0) {
          fcnt++;
  
          refresh = true;
          lcdPrintProg1(readMsg);
          format(6 + 16, avail, 2, 0);
          lcdWrite(15 + 16, fcnt + '0');
          blinkRed(250, 6);
  
          if (fcnt > 3) {
            // not enough bytes read
            vers    = -1;
            logMode = NO_LOGGING;
            storeParams();
            setError(frameError);
          }
          Serial.flush();
        }
        Serial.write('A');
        if (retry > 0) {
          blinkRed(4);
        }
      }
      delay50();
      retry++;
    }
    if (retry >= 600) {
      setOffline(timeoutError);
      return 0;
    }
      
    char prefix = readByte();
    if (prefix == tag) {
      bptr    = buffer;
      *bptr++ = prefix;
  
      int8_t readBytes = 0;
      while (readBytes < len) {
        *bptr++ = readByte();
        readBytes++;
      }
      if (buffer[readBytes] == tag) {
        if (state == 34) {
          vbat = buffer[48];
        }
        else {
          deserialize();
          blinkYellow4();
        }
        if (refresh) {
          refreshDisplay();
        }
        return readBytes;
      }
    }
    // invalid prefix or suffix
    vers    = -1;       // set version invalid to force version request after next start
    logMode = NO_LOGGING;
    storeParams();
    setError(frameError);
  }

  blinkRed(20);
  delay(100);
  return 0;
}
#else
void serRnd(int16_t val MM_DIM) {serialize16(val MM_ACT + random(20) - 9);}

uint8_t serialRead() {
  bptr = buffer;

  boolean v16p = (vers >= V1_6_prb);
  boolean v16  = (vers >= V1_6);
  boolean v17p = (vers >= V1_7_prb);

  serialize8('A');
  int8_t i;
  for (i = 0; i < 3; i++) serRnd(accSmooth[i]);         // acc
  for (i = 0; i < 3; i++) serRnd(gyroData [i]);         // gyro

  serialize16(33);                                      // alt
  if (v16p) {
    serialize16(66);                                    // heading
  }

  int8_t numServos = (v16)?  4 : 2;  
  int8_t numChan   = (v17p)? 8 : 5;  

  for (i = 0; i < numServos; i++) serRnd(servo       [i]); // servos
  for (i = 0; i < 6;         i++) serRnd(motor       [i]); // motor
  for (i = 0; i < numChan;   i++) serRnd(rcHysteresis[i]); // receiver

  if (v17p) {
    serialize8(1);    // 0:nunchukPres, 1:accPres, 2:baroPres, 3:magPres
    serialize8(1);    // 0: accMode, 1:baroMode, 2:magMode
  }
  else {
    serialize8(1);    // nunchukPres
  
    if (v16p) {
      serialize8(0);  // accPres
      serialize8(0);  // baroPres
      serialize8(0);  // magPres
    }
    serialize8(1);    // levelmode
  }
  serialize16(5120);  // cycle time
  
  for (i = 0; i < 2; i++) {
    serRnd(angle[i]);  // roll and pitch angle
  }
  
  serialize8(2);      // config
  for (i = 0; i < 3; i++) {
    serialize8(40);   // pid
    serialize8(30); 
    serialize8(13);
  }
  
  if (v17p) {
    serialize8(140);  // pLevel
    serialize8(45);   // iLevel
  }

  serialize8(v16p? 50 : 100); // dual rate
  serialize8(66);             // expo

  if (!v17p) {
    serialize8(111);  // accStrength
  }

  serialize8(0);      // roll/pitch rate
  serialize8(0);      // yaw rate
  serialize8(0);      // throttle PID rate
  
  if (v17p) {
    for (int8_t i = 0; i < 5; i++) {serialize8(actMode[i]);}
  }
  
  serialize8('A');
    
  deserialize();
  
  if (state != 34) {
    blinkYellow4();
  }
  delay(200);
  
  vbat = 90 + random(30);

  return (bptr - buffer);
}
#endif

void deserialize() {
  bptr = buffer + 1;
  
  boolean v16p = (vers >= V1_6_prb);
  boolean v16  = (vers >= V1_6);
  boolean v17p = (vers >= V1_7_prb);
  
  int8_t i;
  for (i = 0; i < 3; i++) {deserialize16(accSmooth[i]);}
  for (i = 0; i < 3; i++) {deserialize16(gyroData [i]);}
  deserialize16(altitude);

  if (v16p) {
    deserialize16(heading);
  }

  int8_t numServos = (v16) ? 4 : 2;  
  int8_t numChan   = (v17p)? 8 : 5;  
  
  for (i = 0; i < numServos; i++) {deserialize16(servo       [i]);}
  for (i = 0; i < 6;         i++) {deserialize16(motor       [i]);}
  for (i = 0; i < numChan;   i++) {deserialize16(rcHysteresis[i]);}

  if (v17p) {
    int8_t sensorData;  
    deserialize8(sensorData);
    
//    nunchukPres = (sensorData & 1) != 0;
    accPres     = (sensorData & 3);
    baroPres    = (sensorData & 4);
    magPres     = (sensorData & 8);

    deserialize8(sensorData);
    accMode     = (sensorData & 1);
    baroMode    = (sensorData & 2);
    magMode     = (sensorData & 4);
  }
  else {
    deserialize8(accPres); // nunchuk
    if (v16p) {
      deserialize8(i); accPres += 2*i;
      deserialize8(baroPres);
      deserialize8(magPres);
    }  
    deserialize8(accMode);
  }
  
  deserialize16(cycleTime);
  
  for (i=0; i<2; i++) {deserialize16(angle[i]);} 

#ifdef TEST
  bptr++;
#else
  deserialize8(config); 
#endif

  for (i=0; i<3; i++) {
    deserialize8(P8[i]); 
    deserialize8(I8[i]);
    deserialize8(D8[i]);
  }

  if (v17p) {
    deserialize8(P8[3]);
    deserialize8(I8[3]);
  }
  
  deserialize8(rcRate8);
  if (!v16p) {rcRate8 /= 2;}
  deserialize8(rcExpo8); 

  if (!v17p) {deserialize8(accStrength);}
  
  deserialize8(rollPitchRate); 
  deserialize8(yawRate); 
  deserialize8(dynThrPID); 
  
  if (v17p) {
    for (i=0; i<5; i++) {deserialize8(actMode[i]);}
  }
}

void serialWrite() {
  lcd.noBlink();
  
  bptr = buffer;
  
  boolean v17p = (vers >= V1_7_prb);
  
  serialize8('C');
  for (int8_t i=0; i<3; i++) {
    serialize8(P8[i]); 
    serialize8(I8[i]); 
    serialize8(D8[i]);
  }

  if (v17p) {
    serialize8(P8[3]);
    serialize8(I8[3]);
  }

  uint8_t dualRate = (vers == V1_5)? rcRate8 * 2 : rcRate8;
  serialize8(dualRate);
  serialize8(rcExpo8);

  if (!v17p) {serialize8(accStrength);}

  serialize8(rollPitchRate);
  serialize8(yawRate);
  serialize8(dynThrPID);

  if (v17p) {
    for (int8_t i=0; i<5; i++) {serialize8(actMode[i]);}
  }
  
  while (serialAvailable()) {
    Serial.read();
    blinkRed4();
  }
  
  uint8_t len = bptr - buffer;
//  serialWriteBuffer();
  Serial.write(buffer, len);
  Serial.flush();

  lcdPrintProg1(writeMsg);
  format(7 + 16, len, 2, 0);

  blinkYellow500();

  serialRead(); // reload the frame
}

// ----------------

uint8_t readEEPROM (uint16_t adr)              {EEPROM.read (adr);}
void    writeEEPROM(uint16_t adr, uint8_t val) {EEPROM.write(adr, val);}

void storeProfile() {
  getProfileName(lineBuffer); 
  if (enterName(lineBuffer)) {
    storeProfile(lineBuffer);
  }
}

boolean restoreProfile() {
  uint16_t adr  = (profile+1) * 32 + 8; // skip name
  uint8_t* lptr = buffer;
  for (int8_t i = 0; i < 24; i++) {
    *lptr++ = readEEPROM(adr++);
  }
  return extractProfile(buffer);
}
 
boolean extractProfile(uint8_t* lptr) { 
  uint8_t tag = *lptr++;
  boolean v17pre = (tag == 'D');
  boolean v17prb = (tag == 'E');
  boolean v17    = v17pre || v17prb;

  if (v17 || tag == 'C') {
    P8[3]         = 140;
    I8[3]         =  45;
    accStrength   = 100;
    actMode[0]    = 4;
    actMode[1]    = 0;
    actMode[2]    = 0;

    for (int8_t i = 0; i < 3; i++) {
      P8[i]       = *lptr++; 
      I8[i]       = *lptr++; 
      D8[i]       = *lptr++;
    }
    if (v17) {
      P8[3]       = *lptr++;
      I8[3]       = *lptr++;
    }
    rcRate8       = *lptr++;
    rcExpo8       = *lptr++;
    if (!v17) {
      accStrength = *lptr++;
    }
    rollPitchRate = *lptr++;
    yawRate       = *lptr++;
    dynThrPID     = *lptr++;
    if (v17) {
      int n = (v17pre)? 3 : 5;
      for (int8_t i = 0; i < n; i++) {actMode[i] = *lptr++;}
    }

    serialWrite();
    return true;
  }
  return false;
}

void storeProfile(char* name) {
  uint16_t adr  = (profile+1) * 32;
  uint8_t* lptr = buffer;
  int8_t   i    = 0;

  while (name[i] != '\0' && i < 8) { *lptr++ = name[i++]; }
  while (i < 8)                    { *lptr++ = ' ';       }

  boolean v17prb = (vers >= V1_7_prb);
  
  *lptr++   = (v17prb)? 'E' : 'C';
  for (i = 0; i < 3; i++) {
    *lptr++ = P8[i]; 
    *lptr++ = I8[i]; 
    *lptr++ = D8[i];
  }
  if (v17prb) {
    *lptr++ = P8[3];
    *lptr++ = I8[3];
  }

  *lptr++   = rcRate8;
  *lptr++   = rcExpo8;
  if (!v17prb) {
    *lptr++ = accStrength;
  }
  *lptr++   = rollPitchRate;
  *lptr++   = yawRate;
  *lptr++   = dynThrPID;
  if (v17prb) {
    for (i = 0; i < 5; i++) {*lptr++ = actMode[i];}
  }
  
  for (i = 0; i < 32; i++) {
    writeEEPROM(adr++, buffer[i]);
  }
}

void deleteProfile() {
  uint16_t adr = (profile+1) * 32;
  for (int8_t i = 0; i < 32; i++) {
    writeEEPROM(adr++, 0);
  }
}

bool getProfileName(char* name) {
  uint16_t adr = (profile+1) * 32;
  for (int8_t i = 0; i < 8; i++) { name[i] = readEEPROM(adr++); }
  //name[8] = '\0'; 

  char    hdr  = readEEPROM(adr++);
  boolean succ = (hdr == 'C' || hdr == 'D');  
  if (!succ) {
    for (int8_t i = 0; i < 8; i++) { name[i] = '-'; }
  }
  return succ;
} 

// ----------------

boolean restoreParams() {
  byte    magic0 =  readEEPROM(0); // check validity of data by storing some magic numbers as a prefix
  byte    magic1 =  readEEPROM(1);
  uint8_t tv     =  readEEPROM(2);
  boolean succ   = (magic0 == 47 && magic1 == 11 && tv >= 0 && tv < sizeof(frameLength));

  if (succ) {
    vers    = tv;
#ifdef LOGGING
    logMode      = readEEPROM(3);
    if (logMode < NO_LOGGING) {logMode = NO_LOGGING;}
#endif
  }
  delayTime      =       readEEPROM(4);
  backlight      = 255 - readEEPROM(5);
  baudRate       = -1  + readEEPROM(6);
  if (baudRate > 8) {baudRate = 8;} 
  return succ;
}

void storeParams() {
  writeEEPROM(0, 47);  
  writeEEPROM(1, 11);  
  writeEEPROM(2, vers);
#ifdef LOGGING  
  writeEEPROM(3, logMode);
#endif
  writeEEPROM(4, delayTime);
  writeEEPROM(5, 255 - backlight);
  writeEEPROM(6, baudRate + 1);
}

// ----------------

#ifdef LOGGING
boolean fileExists() { return SD.exists(fileName); }

void showLineBlink(prog_char const line[]) {
  lcdPrintProg1(line);
  blinkRed(500, 3);
}

void openLog() {
  int cnt    = 0;
  int offset = 0;
#ifdef LOGVIEW
  if (logMode == LF_LOGVIEW) {
    offset = 500;
  }
#endif

  //  boolean created = false;
  while (!logFile && cnt < 500) {
    formatLogFile(offset + cnt);
    if (!fileExists()) {
      logFile = SD.open(fileName, FILE_WRITE);
    }
    else {
      cnt++;
    }
  }
  if (!logFile) {
    logMode = NO_LOGGING;
    storeParams();
    setError(fileError);
  }
//  state = 0;
  storeParams();

  lcdPrintProg0(logMsg);
//  printFileName();

  if (logMode != LF_LOGVIEW) {
    logFile.write(vers);
    logFile.write(config);
  }
  setBacklight();
  green(true);
}

#ifdef LOGVIEW
byte* toBuffer(byte* outPtr, int16_t value[3]) {return toBuffer(outPtr, value[0]);}

byte* toBuffer(byte* outPtr, int16_t value   ) {
  if (value == 0) {
    *outPtr++ = '0';
  }
  else {
    if (value < 0) {
      value = -value;
      *outPtr++ = '-';
    }
    boolean flag = false;
    int16_t dec  = 10000;
    while (dec > 0) {
      int val = (value / dec) % 10;
      flag |= (val > 0);
      if (flag) {
        *outPtr++ = val + '0';
      }
      dec /= 10;
    }
  }
  *outPtr++ = ';';
  return outPtr;
}

byte* toBufferPwm(byte* outPtr, int16_t value[3], boolean symm) {return toBufferPwm(outPtr, value[0], symm);}

byte* toBufferPwm(byte* outPtr, int16_t value   , boolean symm) {
  if (value != 0) {
    if (symm) {
      value = (value - 1500) / 5; // symmetric [-100..+100]
    }
    else {
      value = (value - 1000) / 10; // percent (0..100)
    }
    outPtr = toBuffer(outPtr, value);
  }
  else {
    *outPtr++ = ';';
  } 
  return outPtr;
}

byte* addPrefix(byte* outPtr, char channel) {
  *outPtr++ = '$';
  *outPtr++ = channel;
  *outPtr++ = ';';
  *outPtr++ = '1' + accMode + baroMode + magMode;  // state
  *outPtr++ = ';';
  outPtr = toBuffer(outPtr, lineNo); // used as timeStamp
  return outPtr;
}

byte* addSuffix(byte* outPtr) {
  *outPtr++ = '0';  // checksum - not used
  *outPtr++ = '\r';
  *outPtr++ = '\n';
  return outPtr;
}
#endif

//int8_t channels[] = { THROTTLE, ROLL, PITCH, YAW, AUX1, AUX2, CAM_ROLL, CAM_PITCH, };
#ifdef LOGVIEW
static uint8_t channelMapping[] = { 3, 0, 1, 2, 4, 5, 6, 7, };
#endif

void doLog() {
  state  = LOG_STATE;
  lineNo = 0;

  while (logMode > NO_LOGGING) {
#ifdef LOGVIEW
    if (logMode == LF_LOGVIEW) {
      deserialize();
  
      byte  outBuffer[160];
      byte* outPtr = outBuffer;
  
      outPtr = addPrefix(outPtr, '1');
      for (uint8_t i = 0; i < 3; i++) { outPtr = toBuffer(outPtr, accSmooth[i]); } // 2 - 4
      for (uint8_t i = 0; i < 3; i++) { outPtr = toBuffer(outPtr, gyroData [i]); } // 5 - 7
      outPtr = toBuffer(outPtr, altitude);                                         // 8
      outPtr = toBuffer(outPtr, heading );                                         // 9
      for (uint8_t i = 0; i < 2; i++) { outPtr = toBuffer(outPtr, angle[i]); }     // 10 - 11
      outPtr = addSuffix(outPtr);
  
      outPtr = addPrefix(outPtr, '2');
      for (uint8_t i = 0; i < 4; i++) { outPtr = toBufferPwm(outPtr, servo[i], true);  } // 1 - 4
      for (uint8_t i = 0; i < 6; i++) { outPtr = toBufferPwm(outPtr, motor[i], false); } // 5 - 10
      outPtr = addSuffix(outPtr);
  
      outPtr = addPrefix(outPtr, '3');
      for (uint8_t i = 0; i < 8; i++) { uint8_t j = channelMapping[i]; outPtr = toBufferPwm(outPtr, rcHysteresis[j], (i>0)); } // 1 - 8
      outPtr = addSuffix(outPtr);
      
      logFile.write(outBuffer, outPtr - outBuffer);
    }
    else
#endif
    {
      logFile.write(buffer, bptr - buffer);
    }
    setBacklight();
  
    lineNo++;
    
    int8_t key = getKeyPressed();
    if (key >= 0) {
      bool flag = confirm(stopLogMsg);
  
      if (flag) {
        logMode = NO_LOGGING;
        break;
      }
      else {
        lcdPrintProg0(logMsg);
//        printFileName();
      }
    }
    if (serialRead() > 0) {
      printFileName();
      format(11 + 16, lineNo, 5, 0);
      blinkYellow4();
      delay(delayTime * 10);
    }
    else {
      setOffline(timeoutError);
    }
  }
  storeParams();
  closeLogFile();
}

int findNextLogFile(int act, boolean ascend, boolean binOnly) {
  int     mod   = (binOnly)? 500 : 1000;
  int     offs  = (ascend)? 1 : mod-1;
  int     cnt   = (act + offs) % mod;
  boolean found = false;
  while (!found && cnt != act) {
    formatLogFile(cnt);
    red(!(cnt & 0x1F));
    found = fileExists();
    if (!found) {
      cnt = (cnt + offs) % mod;
    }
  }
  red(false);
  if (!found) {
    formatLogFile(act);          // check if exactly one logfile exists
    found = fileExists();
  }
 setBacklight();
  return (!found)? -1 : cnt;
}

void formatLogFile(int cnt) {
  if (cnt >= 500) {
    fileName[5] = 'v';    // .lvw - LogView 
    fileName[6] = 'w';    // (Note: '.lov' Files are _not_ in OpenFormat)
  }
  else {
    fileName[5] = 'o';    // .log - binary
    fileName[6] = 'g';
  }
  fileName[2] = '0' +  cnt     % 10;
  fileName[1] = '0' + (cnt/10) % 10;
  fileName[0] = '0' + (cnt/100);
}

void printFileName() {
  lcdPrintProg1(fileMsg);
  for (uint8_t i = 0; i < 3; i++) {
    lcdWrite(fileName[i]);
  }
  lcdPrint(" #");
}

int selectLogfile(int next, boolean binOnly) {
  boolean cancel = false;
  next = findNextLogFile(next, true, binOnly);
  while (next >= 0) {
    boolean valid = false;
    
    uint8_t confId = -1;
    uint8_t fileVers;
    prog_char const* typeStr = configIds;
    if (next < 500) {
      if (openFileRead()) {
        fileVers = logFile.read();
        confId   = logFile.read();
        valid    = (fileVers < sizeof(frameLength));
        if (valid) {
          typeStr = versions[fileVers];
        }
        closeLogFile();
      }
    }
    else {
      valid   = true;
      typeStr = emptyLine;
    }
    
    printFileName();
    lcdPrintProg(10 + 16, typeStr);
    if (confId >= 0 && confId < NUM_CONFIGS) {
      char* chptr = shortConfigIds + (confId*2);
      lcdWrite(*chptr++);
      lcdWrite(*chptr++);
    }

    int8_t key = enterKey();
    if (key == KEY_UP || key == KEY_DOWN) {
      next = findNextLogFile(next, key == KEY_DOWN, binOnly);
    }
    else if (key == KEY_LEFT) {
      next   = -1;
      cancel = true;
    }
    else if (key >= 0 && valid) { 
      break;
    }
  }
  return next;
}

boolean openFileRead() {
  logFile = SD.open(fileName, FILE_READ);

  boolean succ = logFile;
  if (!succ) {
    showLineBlink(fileError);
  }
  green(succ);
  return succ;
}

void closeLogFile() {
  logFile.close();
  setBacklight();
  green(false);
}

void playbackLog() {
  enableAPC220(false);
  Serial.begin(GUI_SERIAL_COM_SPEED);

  int fileNo = 500;
  while (fileNo >= 0) {
    fileNo = selectLogfile(fileNo-1, true);
    if (fileNo >= 0) {
      lineNo = 0;
      if (openFileRead()) {
        uint8_t fileVers = logFile.read();
        uint8_t confId   = logFile.read();
        uint8_t len      = frameLength[fileVers];
        while (logFile.available()) {
          if (getKeyPressed() == KEY_LEFT) {
            showLineBlink(cancelRead);
            break;
          }
          
          printFileName();
          format(11 + 16, lineNo, 5, 0);
          lineNo++;
        
          bptr = buffer;
          uint8_t readBytes = 0;
          while (readBytes < len && logFile.available()) {
            *bptr++ = logFile.read();
            readBytes++;
          }
          setBacklight();
          if (readBytes == len) {
            int     timeout = 0;
            boolean cancel  = false;
            while (!cancel && timeout < 600 && !serialAvailable()) {
              if (!(timeout & 15)) {
                blinkRed(4);
              }
              if (getKeyPressed() == KEY_LEFT) {
                cancel = true;
              }
              delay50();
              timeout++;
            }
            if (cancel || timeout >= 600) {
              showLineBlink((cancel)? cancelRead : timeoutError);
              break;
            }                
            if (readByte() == 'A') {
              blinkYellow4();
              *bptr++ = 'A';
//              serialWriteBuffer();
              Serial.write(buffer, bptr - buffer);
//              Serial.flush();
              delay(delayTime * 10); // in real-time
            }
          }
          else {
            break; // eof
          }
        }
      }
      closeLogFile();
      lcdPrintProg1(emptyLine);
    }
  }
  enableAPC220(true);
  delay50();
  serialBegin();
}
#endif // LOGGING

